package io.tswf.minecraft.metamorphosis.core.transforming

import groovy.transform.CompileStatic
import groovy.transform.Memoized
import groovy.transform.stc.ClosureParams
import groovy.transform.stc.SimpleType
import groovy.util.logging.Slf4j
import io.tswf.groovy.bettergroovy.asm.expansion.generator.ExpansionInjectionContext
import io.tswf.groovy.bettergroovy.asm.expansion.generator.ExpansionInjector
import io.tswf.groovy.bettergroovy.asm.expansion.meta.ExpansionMetadata
import io.tswf.groovy.bettergroovy.asm.expansion.reader.ExpansionMetadataReader
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.TransmutationInjector
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.TransmutationClassReader
import io.tswf.groovy.bettergroovy.asm.transmutation.meta.TransmutationMethodMetadata
import io.tswf.groovy.bettergroovy.asm.view.generation.ViewInjectionContext
import io.tswf.groovy.bettergroovy.asm.view.generation.ViewInjector
import io.tswf.groovy.bettergroovy.asm.view.meta.ViewMetadata
import io.tswf.groovy.bettergroovy.asm.view.reading.ViewMetadataReader
import io.tswf.minecraft.metamorphosis.core.model.HooksInfo
import io.tswf.minecraft.metamorphosis.core.scanning.HooksInfoScanner
import io.tswf.minecraft.metamorphosis.core.side.SidesResolver
import org.objectweb.asm.ClassReader
import org.objectweb.asm.Type
import org.objectweb.asm.tree.ClassNode

import java.util.function.Function
import java.util.stream.Stream

@Slf4j
@CompileStatic
class MetamorphosisTransforms {

    static boolean logTransforms = System.findGlobalProperty("metamorphosis.transforms.logging").orNull()
            ?.toString()?.toLowerCase() == "true"

    @Memoized
    static Set<String> getClassInternalNamesToTransform() {
        return getTypesToTransform()
                .map { it.internalName }
                .toSet()
    }

    @Memoized
    static Set<Type> getTypesToTransform() {
        def viewTargetsStream = getViews().stream()
            .flatMap { it.targets.stream() }

        def expansionTargetsStream = getExpansions().stream()
            .flatMap { it.targets.stream() }

        def transmutationsTargetsStream = getTransmutations().stream()
                .map { it.targetClass }

        return viewTargetsStream
                .joinWith(expansionTargetsStream)
                .joinWith(transmutationsTargetsStream)
                .toSet()
    }

    @Memoized
    static List<ViewMetadata> getViews() {
        def result = getClasses { it.views }
            .map { ViewMetadataReader.readViewMetadata(it) }
            .toList()

        return result
    }

    @Memoized
    static List<ExpansionMetadata> getExpansions() {
        def result =  getClasses { it.expansions }
                .map { ExpansionMetadataReader.readExpansionMetadata(it) }
                .toList()

        return result
    }

    @Memoized
    static List<TransmutationMethodMetadata> getTransmutations() {
        def result = getClasses { it.transmutations }
                .flatMap { TransmutationClassReader.readAllTransmutationsInClass(it).stream() }
                .toList()

        return result
    }

    static Stream<ClassNode> getClasses(
            @ClosureParams(value = SimpleType, options = "io.tswf.minecraft.metamorphosis.core.model.HooksInfo")
            Closure<Set<String>> classesSupplier
    ) {
        return HooksInfoScanner.scanAllHooks()
                .stream()
                .flatMap { Stream.of(it.server, it.client) }
                .flatMap {hooks ->
                    classesSupplier.call(hooks)?.stream()
                            ?.map {
                                def classBytes = hooks.classNameToBytes[it]
                                        .ensureNotNull { new FileNotFoundException("Can't find class bytes for ${it}") }

                                def classNode = new ClassNode()
                                        .tap {
                                            new ClassReader(classBytes).accept(it, ClassReader.SKIP_FRAMES)
                                        }

                                return classNode
                            }

                    ?: Stream.<ClassNode>empty()
                }
    }

    static void transformClassNode(ClassNode targetClass) {
        def targetType = targetClass.asAsmType()

        if (logTransforms) {
            log.info("Start transforming ${targetType.className}...")
        }

        for (def view : getViews()) {
            if (targetType in view.targets) {
                if (logTransforms) {
                    log.info("Injecting view ${view.viewClass.javaName} into ${targetType.className}...")
                }

                ViewInjector.injectViews(new ViewInjectionContext(viewMeta: view, targetClass: targetClass))
            }
        }

        for (def expansion : getExpansions()) {
            if (targetType in expansion.targets) {
                if (logTransforms) {
                    log.info("Injecting expansion ${expansion.expansionClass.javaName} into ${targetType.className}...")
                }

                ExpansionInjector.injectExpansion(new ExpansionInjectionContext(expansionMeta: expansion, targetClass: targetClass))
            }
        }

        for (def transmutation : getTransmutations()) {
            if (targetType == transmutation.targetClass) {
                if (logTransforms) {
                    log.info("Injecting transmutation ${transmutation.transmutationOwner.javaName} into ${targetType.className}...")
                }

                TransmutationInjector.injectTransmutations(Collections.singletonList(transmutation), targetClass)
            }
        }

        if (logTransforms) {
            log.info("Complete transforming ${targetType.className}!")
        }
    }
}
